/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * FAT filesystem
 *
 *   Support for FAT32 added by Dr Stephen Henson
 *
 */

#include "stdlib.h"
#include "fat.h"
#include "mmc.h"

/* read a FAT boot sector */
int fat_read_bootsector(fat_t *fat, unsigned int sector_start, int dtype)
{
  unsigned char buffer[512];
  int d;

  /* read the sector */
  if(!mmc_read(sector_start * 512, buffer, 512)) {
    printb("fat: read error\n");
    return(0);
  }
  if (dtype == FAT_TYPE_UNKNOWN) {

    /* get FAT type */
    fat->type = FAT_TYPE_UNKNOWN;
    if(!__strncmp(&buffer[54], "FAT12   ", 8)) fat->type = FAT_TYPE_12;
    if(!__strncmp(&buffer[54], "FAT16   ", 8)) fat->type = FAT_TYPE_16;
    if(!__strncmp(&buffer[54], "FAT32   ", 8)) fat->type = FAT_TYPE_32;

    if(fat->type == FAT_TYPE_UNKNOWN) {
      printb("fat: unknown fat type\n");
      return(1);
    }
  } else
    fat->type = dtype;

  /* read number of bytes per sector */
  d = buffer[11] | (buffer[12] << 8);
  if(d != 512) {
    printb("fat: sector size mismatch: 0x%x\n", d);
    return(0);
  }

  /* read number of sectors per cluster */
  fat->sectors_per_cluster = buffer[13];

  /* read number of reserved sectors */
  fat->FAT_sector = buffer[14] | (buffer[15] << 8);
  if(fat->FAT_sector < 1) {
    printb("fat: bad # reserved sectors\n");
    return(0);
  }
  fat->FAT_sector += sector_start;
  
  /* read number of FAT */
  fat->number_of_FAT = buffer[16];
  if(fat->number_of_FAT < 1) {
    printb("fat: bad # of FAT\n");
    return(0);
  }
   
  if (fat->type != FAT_TYPE_32)
    /* read number of root directory entries */
    fat->root_size = buffer[17] | (buffer[18] << 8);

  /* read total number of sectors */
  fat->length = buffer[19] | (buffer[20] << 8);

  if (fat->length == 0)
	fat->length = buffer[32] | (buffer[33] << 8)
			| (buffer[34] << 16) | (buffer[35] << 24);

  /* read the number of sectors per fat */
  fat->sectors_per_FAT = buffer[22] | (buffer[23] << 8);

  if (fat->sectors_per_FAT == 0)
	fat->sectors_per_FAT = buffer[36] | (buffer[37] << 8)
			| (buffer[38] << 16) | (buffer[39] << 24);

  /* compute root directory and data area offset */

  fat->data_sector = fat->FAT_sector + fat->number_of_FAT * fat->sectors_per_FAT;
  /* If FAT32 root directory is just like any other: a set of clusters */

  if (fat->type == FAT_TYPE_32) {
    fat->root_cluster = buffer[44] | (buffer[45] << 8)
			| (buffer[46] << 16) | (buffer[47] << 24);
    fat->root_sector = 0;
    fat->label[0] = 0;
  } else {
    /* get volume label */
    memcpy(fat->label, &buffer[43], 11);
    fat->label[11] = 0;
    fat->root_sector = fat->data_sector;
    fat->data_sector += (32 * fat->root_size + 511) / 512;
  }

  fat->root = NULL;

  return(1);
}

int fat_parse_dir_entry(unsigned char *buffer, fat_dir_entry_t *entry,
								fat_t *fat)
{
  /* Zero first byte signals end of directory */
  if(buffer[0] == 00) return (2);
  if (buffer[0] == 0x05 || buffer[0] == 0xe5) return(0);
  /* get 8.3 filename */
  memcpy(entry->filename, &buffer[0], 8);
  memcpy(&entry->filename[9], &buffer[8], 3);
  entry->filename[8] = '.';
  entry->filename[12] = 0;
  entry->attribute = buffer[11];
  entry->cluster = buffer[26] | (buffer[27] << 8);
  if (fat->type == FAT_TYPE_32)
	entry->cluster |= (buffer[20] << 16) | (buffer[21] << 24);
  entry->filesize =  buffer[28]        | (buffer[29] << 8) |
                    (buffer[30] << 16) | (buffer[31] << 24);
  entry->next = NULL; /* empty */
  return(1);
}

fat_dir_entry_t *fat_read_root(fat_t *fat)
{
  int i, dret, num_entries;
  unsigned char *buffer;
  fat_dir_entry_t *root;
  fat_dir_entry_t *tmp;
  cluster_t fat32_root_next;

  if (fat->type == FAT_TYPE_32) {
    /* Allocate memory for single cluster */
    buffer = (unsigned char *) malloc(fat->sectors_per_cluster * 512);
    fat32_root_next = fat->root_cluster;
    if (!fat_read_cluster(fat, fat->root_cluster, buffer))
      return(NULL);
    num_entries = (fat->sectors_per_cluster * 512) / 32;
  } else {
    /* allocate memory for root directory sectors */
    buffer = (unsigned char *) malloc(fat->root_size * 32);
    /* read the root directory */
    if(!mmc_read(fat->root_sector * 512, buffer, fat->root_size * 32))
      return(NULL);
    num_entries = fat->root_size;
  }
  /* parse the root directory */
  root = NULL;
  dret = 2;
  tmp = (fat_dir_entry_t *) malloc(sizeof(fat_dir_entry_t));

  next_root:

  for(i = 0; i < num_entries; i++) {
    dret = fat_parse_dir_entry(buffer + 32 * i, tmp, fat);
    if (dret == 0)
      continue;
    if (dret == 2)
      break;
    if(!(tmp->attribute & FAT_ATTR_LABEL)) { /* used for long filenames */
	/* link the file */
	tmp->next = root;
	root = tmp;
	tmp = (fat_dir_entry_t *) malloc(sizeof(fat_dir_entry_t));
    }
  }

  if (dret != 2 && (fat->type == FAT_TYPE_32)) {
    fat32_root_next = fat_next(fat, fat32_root_next);
    if (fat32_root_next) {
      if (!fat_read_cluster(fat, fat32_root_next, buffer)) {
        return(NULL);
      }
      goto next_root;
    }
  }

  fat->root = root;

  if (!root)
    printb("fat: empty root directory\n");

  return(root);
}

int fat_read_table(fat_t *fat)
{
  unsigned char *buffer;

  /* Just initialize FAT cache for FAT32 */
  if (fat->type == FAT_TYPE_32) {
    fat->table = malloc(512);
    fat->fat_sector = -1;
    return 1;
  }

  /* allocate memory for the table */
  buffer = (unsigned char *) malloc(fat->sectors_per_FAT * 512);

  /* read the allocation table */
  if(!mmc_read(fat->FAT_sector * 512, buffer, fat->sectors_per_FAT * 512)) {
    printb("fat: table read error\n");
    return(0);
  }

  fat->table = buffer;

  return(1);
}

/*
 * For FAT32 the table could be very large so cache a single sector.
 */ 

int fat_read_sector(fat_t *fat, int sector)
{
  if (sector != fat->fat_sector) {
    if(!mmc_read((fat->FAT_sector + sector) * 512, fat->table, 512)) {
      printb("fat: table read error\n");
      return(0);
    }
    fat->fat_sector = sector;
  }

  return(1);
}

int fat_init(fat_t *fat, unsigned int sector_start, int dtype)
{
  /* read the boot sector */
  if(!fat_read_bootsector(fat, sector_start, dtype))
    return(0);

  /* read the cluster allocation table */
  if(!fat_read_table(fat))
    return(0);

  /* read the root directory */
  if(!fat_read_root(fat))
    return(0);

  return(1);
}

cluster_t fat_open_file(fat_t *fat, int *file_size, char const *filename)
{
  fat_dir_entry_t *entry;
  char const *ext = strrchr(filename, '.') + 1;
  int name_length;
  int ext_length;

  /* bug: this is case sensitive and for 8.3 filenames */
  if(strchr(filename, ' '))
    name_length = strchr(filename, ' ') - filename;
  else
    name_length = strchr(filename, '.') - filename;

  if(strchr(ext, ' '))
    ext_length = strchr(ext, ' ') - ext;
  else
    ext_length = 3;

  /* look for the filename in the root dir */
  for(entry = fat->root; entry; entry = entry->next) {
    if(!__strncmp(filename, entry->filename, name_length) &&
       !__strncmp(ext, &entry->filename[9], ext_length)) {
      *file_size = entry->filesize;
      return(entry->cluster);
    }
  }

  return(0);
}

#define FAT_FIRST_CLUSTER 2 /* cluster addresses start at 2 */

cluster_t fat_read_cluster(fat_t *fat, cluster_t cluster, void *buffer)
{
  size_t cluster_size = 512 * fat->sectors_per_cluster;

  if(mmc_read(fat->data_sector * 512 + cluster_size * (cluster - FAT_FIRST_CLUSTER),
	       buffer, cluster_size))
    return(cluster_size);
  else
    return(0);
}

cluster_t fat_next(fat_t *fat, cluster_t cluster)
{
  cluster_t next;
  unsigned int addr;
  unsigned char *table = fat->table;

  /* find the next cluster in the allocation table */
  switch(fat->type) {
  case FAT_TYPE_12:
    addr = (3 * cluster) >> 1;
    if(cluster & 1) /* odd */
      next = (((cluster_t) table[addr + 1]) << 4) |
	     ((((cluster_t) table[addr]) & 0xf0) >> 4);
    else            /* even */
      next = ((cluster_t) table[addr]) |
	     ((((cluster_t) table[addr + 1]) & 0x0f) << 8);
    if(next > 0xff8)
      return(0); /* end of file */
    if(next < 2 || next > 0xff0) {
      printb("fat: unexpected cluster 0x%x\n", next);
      return(0);
    }
    break;
  case FAT_TYPE_16:
    addr = 2 * cluster;
    next = (((cluster_t) table[addr + 1]) << 8) |
	     ((cluster_t) table[addr]);
    if(next > 0xfff8)
      return(0); /* end of file */
    if(next < 2 || next > 0xfff0) {
      printb("fat: unexpected cluster 0x%x\n", next);
      return(0);
    }
    break;
  case FAT_TYPE_32:
    addr = cluster * 4;
    if (!fat_read_sector(fat, addr / 512))
      return 0;
    addr &= 0x1ff;
    next = ((cluster_t) table[addr])
		| (((cluster_t) table[addr + 1]) << 8)
                | (((cluster_t) table[addr + 2]) << 16)
                | (((cluster_t) table[addr + 3]) << 24);
    if (next > 0xffffff8L)
      return(0);
    if(next < 2 || next > 0xffffff0l) {
      printb("fat: unexpected cluster 0x%x\n", next);
      return(0);
    }
    
    break;
  default:
    printb("fat: unsupported type\n");
    return(0);
  }

  return(next);
}

int fat_read_file(fat_t *fat, cluster_t cluster, void *buffer)
{
  void *ptr = buffer;

  while(cluster) {
    ptr += fat_read_cluster(fat, cluster, ptr);
    cluster = fat_next(fat, cluster);
  }

  return(ptr - buffer);
}

