/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * framebuffer and console routines
 *
 */

#include "fb.h"
#include "font_6x11.h"     /* 6x11 font by rthelen */
#include "stdlib.h"
#include "gpio.h"
#include "model.h"

/* the framebuffer */
static fb_pix_t * const fb = (fb_pix_t *) FBADDR;
static int fb_width;
static int fb_height;

#define OMAP_LCDC_BASE			0xfffec000
#define OMAP_LCDC_CONTROL		(OMAP_LCDC_BASE + 0x00)
#define OMAP_LCDC_TIMING0		(OMAP_LCDC_BASE + 0x04)
#define OMAP_LCDC_TIMING1		(OMAP_LCDC_BASE + 0x08)
#define OMAP_LCDC_TIMING2		(OMAP_LCDC_BASE + 0x0c)
#define OMAP_LCDC_STATUS		(OMAP_LCDC_BASE + 0x10)
#define OMAP_DMA_LCD_BASE		0xfffee300
#define OMAP_DMA_LCD_CCR		(OMAP_DMA_LCD_BASE + 0xc2)
#define OMAP_DMA_LCD_CTRL		(OMAP_DMA_LCD_BASE + 0xc4)

#define OMAP_LCDC_CTRL_LCD_EN		(1 << 0)
#define OMAP_LCDC_STAT_DONE		(1 << 0)

/* turn off the lcd controller */
void lcd_off(void)
{
  /* disable controller */
  REG32(OMAP_LCDC_CONTROL) &= ~OMAP_LCDC_CTRL_LCD_EN;

  /* wait for end of frame */
  while(!(REG32(OMAP_LCDC_STATUS) & OMAP_LCDC_STAT_DONE));

  /* turn off DMA */
  REG16(OMAP_DMA_LCD_CCR)  &= ~(1 << 7);
  REG16(OMAP_DMA_LCD_CTRL) &= ~(1 << 8);
}

void fb_init(void)
{
  switch (get_model()) {
  case MODEL_TYPHOON:
    fb_width = 176;
    fb_height = 220;
    break;
  case MODEL_TORNADO:
    fb_width = 240;
    fb_height = 320;
    break;
  default:
    /* unknown resolution, try to read it from the LCD controller registers */
    fb_width  = (REG32(OMAP_LCDC_TIMING0) & 0x3ff) + 1;
    fb_height = (REG32(OMAP_LCDC_TIMING1) & 0x3ff) + 1;
    break;
  }
}

void dump_fb_timing_info(void)
{
  printb("T0: %x\n", REG32(OMAP_LCDC_TIMING0));
  printb("T1: %x\n", REG32(OMAP_LCDC_TIMING1));
  printb("T2: %x\n", REG32(OMAP_LCDC_TIMING2));
}

void clear_fb(fb_pix_t color)
{
  int i;
  
  for(i = 0; i < fb_width * fb_height; i++)
    fb[i] = color;
}

void draw_char(int x, int y, int c, fb_pix_t color)
{
  int i, j;
  int w, h;

  w = FONTWIDTH;
  h = FONTHEIGHT;
  if(c > 255) return;
  if(x < 0) return;
  if(y < 0) return;
  if(x + w > fb_width ) w = fb_width  - x;
  if(y + h > fb_height) h = fb_height - y;

  for(j = 0; j < h; j++) 
    for(i = 0; i < w; i++)
      if(fontdata[c * FONTHEIGHT + j] & (1 << (7 - i)))
	fb[(y + j) * fb_width + (x + i)] = color;
}

/* 'console' stuff */
#define TEXTCOLOR RGB16(0xff,0xff,0xff)
#define BGCOLOR   RGB16(0x00,0x00,0x00)

static int cursor_x, cursor_y;

void console_init()
{
  cursor_x = 0;
  cursor_y = 0;
  clear_fb(BGCOLOR);
}

int putchar(int c)
{
  if(cursor_x + FONTWIDTH > fb_width || c == '\n') { /* new line */
    cursor_y += FONTHEIGHT;
    cursor_x = 0;
  }

  if(cursor_y + FONTHEIGHT > fb_height) { /* clear page */
    cursor_y = 0;
    cursor_x = 0;
    clear_fb(BGCOLOR);
  }

  if(c == '\r')
    cursor_x = 0;

  if(c != '\r' && c != '\n') { /* standard char */
    draw_char(cursor_x, cursor_y, c, TEXTCOLOR);
    cursor_x += FONTWIDTH;
  }

  return(c);
}
