/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *                  Thomas Cougnard <thomas.cougnard@free.fr>
 *
 * bootloader
 *
 */

#include "stdlib.h"
#include "fb.h"
#include "mmc.h"
#include "disk.h"
#include "fat.h"
#include "tag.h"
#include "led.h"
#include "pin.h"
#include "gpio.h"
#include "model.h"
#include "crc.h"

extern void enter_linux(void *address, unsigned int mach_type, void *taglist);
extern void irq_disable(void);
extern void dead_start();
extern void dead_end();

extern void *end;
extern void *start;

#define IMAGE_NAME     "IMAGE.BIN"
#define INITRD_NAME    "INITRD.BIN"
#define LINUX_VIRT     0x8c008000 /* virtual address where linux is loaded */
#define LINUX_PHYS     0x10008000 /* physical address where linux is loaded */
#define TAGLIST_VIRT   0x8c000100 /* virtual address of boot parameters */
#define TAGLIST_PHYS   0x10000100 /* physical address of boot parameters */
#define TAGLIST_SIZE   0x00001000 /* this should be largely enough */
#define INITRD_PHYS    0x10400000 /* physical address of initrd (4Mo) */

/* from http://www.arm.linux.org.uk/developer/machines */
#define MACH_TYPHOON 689 

#define DELAY 100000 /* wait for 10 sec if power button pressed */

int main(void)
{
  int i;
  int found = 0;
  unsigned char buffer[512*32];
  pentry_t ptable[20];
  int pcount = 0;
  fat_t fat;
  fat_dir_entry_t *entry;
  cluster_t fd;
  int size, file_size;
  int dtype;

  /* initialize the led */
  led_init();
  led_set(LED_BLUE);

  /* initialize the console */
  fb_init();
  console_init();

  printb("EOL ");
  printb("%p - %p\n", start, end);
  print_model_info();

  /* mux pins */
  pin_init();

  /* read the SD card */
  mmc_reset();

  /* load the partition table */
  mmc_read(0, buffer, 512);
  pcount += read_ptable(buffer, NULL, ptable);
  printb("SD card partition table:\n");

  dtype = FAT_TYPE_UNKNOWN;

  /* look for the first valid FAT partition */
  for(i = 0; i < pcount; i++) {
    print_pentry(&ptable[i]);

    switch(ptable[i].type) {
    case 0x00: /* empty */
      break;
    case 0x01: /* FAT12 */
      found = 1;
      break;
    case 0x05: /* extended */
      /* read the first sector of the extended partition */
      if (pcount <= 16) {
	mmc_read(ptable[i].sector_start * 512, buffer, 512);
	pcount += read_ptable(buffer, &ptable[i], &ptable[pcount]);
      } else
	printb("too many extended partitions\n");
      break;
    case 0x0b: /* FAT32 */
    case 0x0c: /* FAT32 (LBA) */
      dtype = FAT_TYPE_32;
      found = 1;
      break;
    case 0x04: /* FAT16 <32M */
    case 0x06: /* FAT16 */
    case 0x0e: /* FAT16 (LBA) */
      found = 1;
      break;
    default:
      printb("partition %x: type %x unknown\n", i, ptable[i].type);
      break;
    }
    if(found) break;
  }
  if(!found) {
    printb("No valid partition found\n");
    return(1);
  }

  /* read the FAT */
  if(!fat_init(&fat, ptable[i].sector_start, dtype)) {
    printb("FAT init failure\n");
    return(1);
  }

  /*  
  printb("FAT%d %s %dK\n", fat.type, fat.label,
	     fat.sectors_per_cluster * 512 / 1024);
  */

  /* list the files in the FAT root directory */
  for(entry = fat.root; entry; entry = entry->next)
    printb(" %s %d 0x%x\n", entry->filename, entry->filesize, entry->cluster);

  /* load linux */
  fd = fat_open_file(&fat, &file_size, IMAGE_NAME);
  if(!fd) {
         printb("Could not open %s\n", IMAGE_NAME);
         die();
  }

  size = fat_read_file(&fat, fd, (void *) LINUX_VIRT);
  printb("Kernel loaded... (%dKb)\n", size / 1024);
  printb("CRC: %u\n", cksum((void *) LINUX_VIRT, file_size));

  /* load initrd */
  size = 0;
  fd = fat_open_file(&fat, &file_size, INITRD_NAME);
  if(fd) {
	size = fat_read_file(&fat, fd, (void *) INITRD_PHYS);

 	printb("Ramdisk loaded... (%dKb)\n", size / 1024);
	printb("CRC: %u\n", cksum((void *) INITRD_PHYS, file_size));
  }

  /* check if power button pressed */
  gpio_dir(GPIO_POWER_BUTTON, 1);

  if(!gpio_in(GPIO_POWER_BUTTON)) {
	  printb("Power pressed, reading CID...\n");

	  /* get the card id */
	  if(mmc_get_cid(buffer)) {
		  printb("cardid: ");
		  for(i = 0; i < 16; i++) {
			  printb("%x", buffer[i] >> 4);
			  printb("%x", buffer[i] & 0xf);
		  }
	  }

	  /* wait */
	  mdelay(DELAY);
  }

  /* prepare the tag list (parameters for linux) */
  prepare_tag_list((void *) TAGLIST_VIRT, (void *) INITRD_PHYS, size);

  printb("Entering linux...\n");

  led_set(LED_GREEN | LED_RED);

  /* turn off the lcd */
  /* XXX: You may want to comment that to keep the LCD on for early debugging,
     but the Linux driver may not initialize properly if this is removed. */
  lcd_off();

  led_set(LED_RED);

  /* enter the gate, Charon awaits */
  enter_linux((void *) LINUX_PHYS, MACH_TYPHOON, (void *) TAGLIST_PHYS);

  /* stop here */
  while(1);
  return(0);
}
