/* merge a disk image with a bootloader to create a bootable disk image */
/* the resulting image structure is as follows:
     0x000 - 0x1b0       : bootloader header (boot.img)
     0x1b0 - 0x200       : partition table   (disk.img)
     0x200 - boot.end    : bootloader code   (boot.img)
     boot.end - disk.end : disk data         (disk.img) */

#include <stdio.h>
#include "disk.h"

int main(int argc, char *argv[])
{
  char *diskname, *bootname;
  FILE *disk, *boot;
  int length, i, boot_length;
  pentry_t ptable[4];
  char buffer[0x200];

  if(argc != 3) {
    fprintf(stderr, "usage: %s disk.img boot.img\n", argv[0]);
    fprintf(stderr, "\tmakes disk.img bootable using boot.img\n");
    return(1);
  }
  
  diskname = argv[1];
  bootname = argv[2];

  /* open files */
  if((disk = fopen(diskname, "r+b")) == NULL) {
    fprintf(stderr, "could not open %s in read/write mode: ", diskname);
    perror("");
    return(2);
  }

  if((boot = fopen(bootname, "rb")) == NULL) {
    fprintf(stderr, "could not open %s in read mode: ", bootname);
    perror("");
    return(3);
  }

  fseek(boot, 0, SEEK_END);
  boot_length = ftell(boot);
  fseek(boot, 0, SEEK_SET);
  
  /* copy boot image header to disk */
  fread(buffer, 0x200, 1, disk);
  /* Check partition table */
  if (!read_ptable((unsigned char *)buffer, NULL, ptable)) {
   fprintf(stderr, "Error Reading Partition Table\n");
   return(4);
  }
  printf("Partition Table:\n");
  for (i = 0; i < 3; i++) {
    if (ptable[i].sector_length) {
      print_pentry(&ptable[i]);
      if ((ptable[i].sector_start * 512) < boot_length) {
        fprintf(stderr, "Not enough room for boot image\n");
        return(5);
      }
    }
  }
  fread(buffer, 0x1b0, 1, boot);
  fseek(disk, 0, SEEK_SET);
  fwrite(buffer, 0x200, 1, disk);
  fread(buffer, 0x200 - 0x1b0, 1, boot);

  /* copy boot image content to disk */
  while((length = fread(buffer, 1, 0x200, boot)) > 0)
    fwrite(buffer, length, 1, disk);

  /* close files */
  fclose(disk);
  fclose(boot);

  return(0);
}

/* decode C/H/S address */
void chs_decode(chs_t *chs, unsigned char *addr)
{
  chs->h = addr[0];
  chs->s = addr[1] & 0x3f;
  chs->c = ((addr[1] & 0xc0) << 2) | addr[2];
}

/* read a partition table entry */
void read_pentry(unsigned char *buffer, pentry_t *pentry)
{
  pentry->boot = buffer[0];
  chs_decode(&pentry->begin, &buffer[1]);
  pentry->type = buffer[4];
  chs_decode(&pentry->end, &buffer[5]);
  pentry->sector_start   = buffer[ 8];
  pentry->sector_start  |= buffer[ 9] <<  8;
  pentry->sector_start  |= buffer[10] << 16;
  pentry->sector_start  |= buffer[11] << 24;
  pentry->sector_length  = buffer[12];
  pentry->sector_length |= buffer[13] <<  8;
  pentry->sector_length |= buffer[14] << 16;
  pentry->sector_length |= buffer[15] << 24;
}

/* read the partition table from the bootsector */
int read_ptable(unsigned char bootsector[], pentry_t *parent, pentry_t *ptable)
{
  /* check for the signature at offset 510 */
  if(bootsector[510] != 0x55 || bootsector[511] != 0xaa) {
    fprintf(stderr, "read_ptable: no signature\n");
    return(0);
  }

  /* partition descriptors are at offset 446, 462, 478 and 494 */
  if (!parent) {
    /* MBR */
    read_pentry(&bootsector[446], &ptable[0]);
    read_pentry(&bootsector[462], &ptable[1]);
    read_pentry(&bootsector[478], &ptable[2]);
    read_pentry(&bootsector[494], &ptable[3]);
    return 4;
  } else {
    /* EMBR */
    read_pentry(&bootsector[446], &ptable[0]);
    read_pentry(&bootsector[462], &ptable[1]);
    ptable[0].sector_start += parent->sector_start;
    ptable[1].sector_start += parent->sector_start;
    return 2;
  }
}

/* print a partition entry */
void print_pentry(pentry_t *pentry)
{
  printf("%c ", pentry->boot ? 'B' : '-');
  printf("%x ", pentry->type);

  printf("%d ", pentry->sector_start);
  printf("%d ", pentry->sector_length);

  printf("%x/%x/%x ", pentry->begin.c, pentry->begin.h, pentry->begin.s);
  printf("\n");
}
