/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * miniSD I/O
 *
 */
#include "stdlib.h"
#include "mmc.h"

/* initialize data transfer parameters */
void mmc_data_init(int sector_size, int sector_count)
{
  MMC_WRITE(BUF, 0x1f00); /* buffer full at 31, empty at 0, dma disabled */
  MMC_WRITE(BLEN, sector_size - 1);
  MMC_WRITE(NBLK, sector_count - 1);
}

/* send a command to the mmc controller */
int mmc_send_command(int command, int arg)
{
  MMC_WRITE(CTO, 200); /* timeout of 200 cycles */

  /* send the argument */
  MMC_WRITE(ARGL, arg & 0xffff);
  MMC_WRITE(ARGH, arg >> 16);

  /* send the command */
  MMC_WRITE(CMD, command);
  while(!(MMC_READ(STAT) & (MMC_STAT_CMD_TOUT | MMC_STAT_END_OF_CMD | MMC_STAT_CARD_ERR)));
  udelay(10);

  if(MMC_READ(STAT) & MMC_STAT_CMD_TOUT) {
    /* timeout */
//  printb("timeout\n");
    MMC_WRITE(STAT, MMC_STAT_CMD_TOUT); /* ack */
    return(0); 
  }
  if(MMC_READ(STAT) & MMC_STAT_CARD_ERR) {
    /* card error: read response register */
    int response = MMC_READ(RSP6) | (MMC_READ(RSP7) << 16);
    /* card error */
     MMC_WRITE(STAT, MMC_STAT_CARD_ERR); /* ack */
//    printb("mmc_send_command: card error %x\n", response);
//    printb("mmc: ARGL = %x\n", MMC_READ(ARGL));
//    printb("mmc: ARGH = %x\n", MMC_READ(ARGH));
    return(0);
  }

  MMC_WRITE(STAT, MMC_STAT_END_OF_CMD); /* ack */
  return(1);
}

/* get the MMC unique card id */
int mmc_get_cid(char cid[16])
{
    int retry = 10;
    unsigned short *data = (unsigned short *) cid;

    if(!mmc_send_command(MMC_CMD_GO_IDLE, 0))
	    return(0);
    do {
	    mmc_send_command(MMC_CMD_APP_CMD, 0);
	    if(mmc_send_command(SD_CMD_APP_OP_COND, 0xff8000)) break;
	    udelay(100);
    } while(retry--);

    if(!retry) {
	    printb("timeout waiting for SD card\n");
	    return(0);
    }

    if(!mmc_send_command(MMC_CMD_SEND_CID, 0)) {
	    printb("could not read SD card CID\n");
	    return(0);
    }

    data[0] = MMC_READ(RSP0);
    data[1] = MMC_READ(RSP1);
    data[2] = MMC_READ(RSP2);
    data[3] = MMC_READ(RSP3);
    data[4] = MMC_READ(RSP4);
    data[5] = MMC_READ(RSP5);
    data[6] = MMC_READ(RSP6);
    data[7] = MMC_READ(RSP7);

    return(1);
}

/* read some data from mmc controller buffer in PIO mode */
/* count is in sectors of 512 bytes */
int mmc_read_data(void *buffer, int count)
{
  int i, j, n;
  unsigned short *data = (unsigned short *) buffer;

  if(count > 0) {
    n = count * SECTOR_SIZE / 64; /* read by chunks of 64 bytes */

    for(j = 0; j < n; j++) {

      while(!(MMC_READ(STAT) & MMC_STAT_A_FULL)); /* wait for data */
      MMC_WRITE(STAT, MMC_STAT_A_FULL); /* ack */
      for(i = 0; i < 32; i++) /* 32 short words = 64 bytes */
	*data++ = MMC_READ(DATA);
    }
  }

  while(!(MMC_READ(STAT) & MMC_STAT_END_OF_DATA)); /* wait for data */
  MMC_WRITE(STAT, MMC_STAT_END_OF_DATA); /* ack */

  return(1);
}

/* read 512-byte sectors from the MMC */
int mmc_read_sectors(int sector_offset, void *buffer, size_t sector_count)
{
  if(sector_count <= 0) return(1);

  mmc_data_init(SECTOR_SIZE, sector_count);

  if(sector_count > 1) { /* multiple block request */
    mmc_send_command(MMC_CMD_READ_MULTIPLE_BLOCK, sector_offset * SECTOR_SIZE);
    mmc_read_data(buffer, sector_count);
    mmc_send_command(MMC_CMD_STOP_TRANSMISSION, 0);
  } else {               /* single block request   */
    mmc_send_command(MMC_CMD_READ_SINGLE_BLOCK, sector_offset * SECTOR_SIZE);
    mmc_read_data(buffer, sector_count);
  }
  
  return(1);
}

/* read data from the MMC */
int mmc_read(int offset, void *buffer, size_t length)
{
  unsigned char *data  = (unsigned char *) buffer;
  unsigned char sector_data[SECTOR_SIZE];
  int start_length = (SECTOR_SIZE - (offset & SECTOR_MASK)) & SECTOR_MASK;
  int end_length = (offset + length) & SECTOR_MASK;
  int start_sector = offset >> SECTOR_SHIFT;
  int end_sector = (offset + length) >> SECTOR_SHIFT;
  int sector_count = (length - start_length) >> SECTOR_SHIFT;

  if(start_length) {
    if(!mmc_read_sectors(start_sector, sector_data, 1))
      return(0);
    memcpy(data, sector_data + (offset & SECTOR_MASK), start_length);
    data += start_length;
    start_sector++;
  }

  if(!mmc_read_sectors(start_sector, data, sector_count))
    return(start_length);
  data += sector_count * SECTOR_SIZE;

  if(end_length) {
    if(!mmc_read_sectors(end_sector, sector_data, 1))
      return(sector_count * SECTOR_SIZE + start_length);
    memcpy(data, sector_data, end_length);
  }

  return(length);
}

/* reset the MMC controller */
int mmc_reset()
{
  /* clear the status register */
  MMC_WRITE(STAT, 0xffff);

  /* send initialisation stream/data abort command */
  mmc_send_command(MMC_CMD_RESET, 0);

  return(MMC_READ(STAT));
}
