/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * Prepare Linux boot parameters
 *
 */

#include "stdint.h"
#include "setup.h" /* kernel interface */
#include "board.h" /* kernel interface (OMAP) */
#include "stdlib.h"
#include "model.h"

#define MKDEV(ma, mi) (((ma) << 8) | (mi)) /* old-style major/minor */

#define omap_tag_next(entry) ((struct omap_board_config_entry *) \
  ((void *) (entry) + sizeof(*entry) + (entry)->len))

/* fill in the OMAP specific information for the HTC Typhoon */
int prepare_omap_tag_list(void *omaplist_addr)
{
  struct omap_board_config_entry *entry;
  struct omap_boot_reason_config *reason;
  struct omap_lcd_config *lcd;
  char const *lcd_panel;

  entry = (struct omap_board_config_entry *) omaplist_addr;

  /* reason tag */
  entry->tag = OMAP_TAG_BOOT_REASON;
  entry->len = sizeof(struct omap_boot_reason_config);
  reason = (struct omap_boot_reason_config *) entry->data;
  memcpy(reason->reason_str, "eol: cold", sizeof("eol: cold") + 1);
  entry = omap_tag_next(entry);

  /* LCD panel tag */
  entry->tag = OMAP_TAG_LCD;
  entry->len = sizeof(struct omap_lcd_config);
  lcd = (struct omap_lcd_config *) entry->data;
  memcpy(lcd->ctrl_name, "internal", sizeof("internal") + 1);

  lcd_panel = NULL;
  if (get_model() == MODEL_TYPHOON) {
    switch (get_lcd_id()) {
    case TYPHOON_LCD_TOPPOLY:
      lcd_panel = "typhoon-toppoly";
      break;
    case TYPHOON_LCD_HUNDAI:
      lcd_panel = "typhoon-hundai";
      break;
    case TYPHOON_LCD_SONY:
      lcd_panel = "typhoon-sony";
      break;
    }
  } else if (get_model() == MODEL_TORNADO) {
    switch (get_lcd_id()) {
    case TORNADO_LCD_SONY:
      lcd_panel = "tornado-sony";
      break;
    case TORNADO_LCD_TOPPOLY:
      lcd_panel = "tornado-toppoly";
      break;
    }
  }

  if (lcd_panel)
    memcpy(lcd->panel_name, lcd_panel, strlen(lcd_panel) + 1);
  else
    memset(lcd->panel_name, 0, sizeof(lcd->panel_name));

  entry = omap_tag_next(entry);

  return ((sizeof(struct tag_header) + ((void *) entry - omaplist_addr)) >> 2);
}

/* fill in a boot parameter tag list for the HTC Typhoon */
void  prepare_tag_list(void *taglist_addr, void *initrd, int initrd_size)
{
  struct tag *list = (struct tag *) taglist_addr;

  /* core tag (start of list) */
  list->hdr.size = tag_size(tag_core);
  list->hdr.tag = ATAG_CORE;
  list->u.core.flags = 1;                 /* mount root read-only */
  list->u.core.pagesize = 4 * 1024;       /* unused */
  list->u.core.rootdev = MKDEV(1,0);      /* initrd (/dev/ram0) */
  list = tag_next(list);

  /* mem tags */
  list->hdr.size = tag_size(tag_mem32);
  list->hdr.tag = ATAG_MEM;
  list->u.mem.start = 0x10000000;         /* SDRAM base */
  list->u.mem.size  = 32 * 1024 * 1024;   /* SDRAM size  - GSM reserved */
  list = tag_next(list);

  if(initrd && (initrd_size > 0)) {

    /* ramdisk tag */
    list->hdr.size = tag_size(tag_ramdisk);
    list->hdr.tag = ATAG_RAMDISK;
    list->u.ramdisk.flags = 1;            /* load */
    list->u.ramdisk.size = 4096;
    list->u.ramdisk.start = 0;            /* unused */
    list = tag_next(list);

    /* initrd2 tag */
    list->hdr.size = tag_size(tag_initrd);
    list->hdr.tag = ATAG_INITRD2;
    list->u.initrd.start = (u32) initrd; /* initrd start */
    list->u.initrd.size = initrd_size;   /* initrd size  */
    list = tag_next(list);
  }

  /* OMAP tags */
  list->hdr.tag = ATAG_BOARD;
  list->hdr.size = prepare_omap_tag_list(&list->u); /* words */
  list = tag_next(list);

  /* none tag (end of list) */
  list->hdr.size = 0;
  list->hdr.tag = ATAG_NONE;
}
